import os
import time
import threading
from typing import Callable
from cobweb.db import RedisDB, ApiDB
from cobweb.utils import check_pause
from cobweb.base import Queue, Seed, logger
from cobweb.constant import LogTemplate
from .scheduler import Scheduler
use_api = bool(int(os.getenv("REDIS_API", 0)))


class RedisScheduler(Scheduler):

    def __init__(
            self,
            task,
            project,
            stop: threading.Event,
            pause: threading.Event,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            register: Callable
    ):
        super().__init__(task, project, stop, pause, new, todo, done, upload, register)
        self.todo_key = "{%s:%s}:todo" % (project, task)
        self.done_key = "{%s:%s}:done" % (project, task)
        self.fail_key = "{%s:%s}:fail" % (project, task)
        self.heartbeat_key = "heartbeat:%s_%s" % (project, task)
        self.speed_control_key = "speed_control:%s_%s" % (project, task)
        self.reset_lock_key = "lock:reset:%s_%s" % (project, task)
        self.redis_queue_empty_event = threading.Event()
        self.db = ApiDB() if use_api else RedisDB()

    @check_pause
    def reset(self):
        """
        检查过期种子，重新添加到redis缓存中
        """
        reset_wait_seconds = 30
        if self.db.lock(self.reset_lock_key, t=60):

            _min = -int(time.time()) + self.seed_reset_seconds
            self.db.members(self.todo_key, 0, _min=_min, _max="(0")
            self.db.delete(self.reset_lock_key)

        time.sleep(reset_wait_seconds)

    @check_pause
    def schedule(self):
        """
        调度任务，获取redis队列种子，同时添加到doing字典中
        """
        if not self.db.zcount(self.todo_key, 0, "(1000"):
            time.sleep(self.scheduler_wait_seconds)
        elif self.todo.length >= self.todo_queue_size:
            time.sleep(self.todo_queue_full_wait_seconds)
        else:
            members = self.db.members(
                self.todo_key, int(time.time()),
                count=self.todo_queue_size,
                _min=0, _max="(1000"
            )
            for member, priority in members:
                seed = Seed(member, priority=priority)
                self.working_seeds[seed.to_string] = seed.params.priority
                self.todo.push(seed)

    @check_pause
    def insert(self):
        """
        添加新种子到redis队列中
        """
        new_seeds = {}
        del_seeds = set()
        status = self.new.length < self.new_queue_max_size
        for _ in range(self.new_queue_max_size):
            seed_tuple = self.new.pop()
            if not seed_tuple:
                break
            seed, new_seed = seed_tuple
            new_seeds[new_seed.to_string] = new_seed.params.priority
            del_seeds.add(seed)
        if new_seeds:
            self.db.zadd(self.todo_key, new_seeds, nx=True)
        if del_seeds:
            self.done.push(list(del_seeds))
        if status:
            time.sleep(self.new_queue_wait_seconds)

    @check_pause
    def refresh(self):
        """
        刷新doing种子过期时间，防止reset重新消费
        """
        if self.working_seeds:
            refresh_time = int(time.time())
            seeds = {k:-refresh_time - v / 1000 for k, v in self.working_seeds.items()}
            self.db.zadd(self.todo_key, item=seeds, xx=True)
        time.sleep(3)

    @check_pause
    def delete(self):
        """
        删除队列种子，根据状态添加至成功或失败队列，移除doing字典种子索引
        """
        seed_list = []
        status = self.done.length < self.done_queue_max_size

        for _ in range(self.done_queue_max_size):
            seed = self.done.pop()
            if not seed:
                break
            seed_list.append(seed.to_string)

        if seed_list:

            self.db.zrem(self.todo_key, *seed_list)
            self.remove_working_seeds(seed_list)

        if status:
            time.sleep(self.done_queue_wait_seconds)

    def run(self):
        start_time = int(time.time())

        self.register(self.reset, tag="scheduler")
        self.register(self.insert, tag="scheduler")
        self.register(self.delete, tag="scheduler")
        self.register(self.refresh, tag="scheduler")
        self.register(self.schedule, tag="scheduler")

        while not self.stop.is_set():
            working_count = len(self.working_seeds.keys())
            memory_count = self.db.zcount(self.todo_key, "-inf", "(0")
            todo_count = self.db.zcount(self.todo_key, 0, "(1000")
            all_count = self.db.zcard(self.todo_key)
            
            if self.is_empty():
                if self.pause.is_set():
                    execute_time = int(time.time()) - start_time
                    if not self.task_model and execute_time > self.before_scheduler_wait_seconds:
                        logger.info("Done! ready to close thread...")
                        self.stop.set()
                    elif todo_count:
                        logger.info(f"Recovery {self.task} task run！todo seeds count: {todo_count}, queue length: {all_count}")
                        self.pause.clear()
                        # self.execute()
                    else:
                        logger.info("pause! waiting for resume...")
                else:
                    if all_count:
                        logger.info(f"todo seeds count: {todo_count}, queue length: {all_count}")
                        self.pause.is_set()
                    else:
                        logger.info("Done! pause set...")
                        self.pause.clear()
            else:
                if self.pause.is_set():
                    self.pause.clear()
                logger.info(LogTemplate.launcher_pro_polling.format(
                    task=self.task,
                    doing_len=working_count,
                    todo_len=self.todo.length,
                    done_len=self.done.length,
                    redis_seed_count=all_count,
                    redis_todo_len=todo_count,
                    redis_doing_len=memory_count,
                    upload_len=self.upload.length,
                ))

            time.sleep(30)

        logger.info("Done! Ready to close thread...")




import time
import threading
import traceback
from inspect import isgenerator
from typing import Callable
from urllib.parse import urlparse
from requests import Response as Res

from cobweb import setting
from cobweb.constant import DealModel, LogTemplate
from cobweb.base import (
    Seed,
    Queue,
    BaseItem,
    Request,
    Response,
    logger
)
from cobweb.utils import LoghubDot, check_pause


class Distributor(threading.Thread):

    def __init__(
            self,
            task: str,
            project: str,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            register: Callable,
            stop: threading.Event,
            pause: threading.Event,
            SpiderCrawler
    ):
        super().__init__()
        self.task = task
        self.project = project
        self.stop = stop
        self.pause = pause

        self.new = new
        self.todo = todo
        self.done = done
        self.upload = upload
        self.register = register

        self.time_sleep = setting.SPIDER_TIME_SLEEP
        self.thread_num = setting.SPIDER_THREAD_NUM
        self.max_retries = setting.SPIDER_MAX_RETRIES
        self.record_failed = setting.RECORD_FAILED_SPIDER
        self.loghub_dot = LoghubDot()  # todo: 解偶

        self.Crawler = SpiderCrawler

        logger.debug(f"Distribute instance attrs: {self.__dict__}")

    def distribute(self, item, seed, _id: int):
        if isinstance(item, Request):
            seed.params.start_time = time.time()
            self.process(item=item, seed=seed, callback=self.Crawler.download, _id=1)
        elif isinstance(item, Response):
            if _id == 2:
                raise TypeError("parse function can't yield a Response instance")
            dot = isinstance(item.response, Res)
            self.spider_logging(seed, item, dot=dot)
            self.process(item=item, seed=seed, callback=self.Crawler.parse, _id=2)
        elif isinstance(item, BaseItem):
            self.upload.push(item)
        elif isinstance(item, Seed):
            self.new.push((seed, item), direct_insertion=True)
        elif isinstance(item, str) and item == DealModel.poll:
            self.todo.push(seed)
        elif isinstance(item, str) and item == DealModel.done:
            self.done.push(seed)
        elif isinstance(item, str) and item == DealModel.fail:
            seed.params.retry += 1
            if seed.params.retry < self.max_retries:
                self.todo.push(seed)
            else:
                if record_failed := self.record_failed:
                    try:
                        response = Response(seed, "failed", max_retries=True)
                        self.process(response, seed, self.Crawler.parse, _id=2)
                    except:
                        record_failed = False
                if not record_failed:
                    self.done.push(seed)
        else:
            raise TypeError("yield value type error!")

    def process(self, item, seed, callback, _id: int):
        result_iterators = callback(item)
        if not isgenerator(result_iterators):
            raise TypeError(f"{callback.__name__} function isn't a generator!")
        for result_item in result_iterators:
            self.distribute(result_item, seed, _id)

    @check_pause
    def spider(self):
        if seed := self.todo.pop():
            try:
                self.process(item=seed, seed=seed, callback=self.Crawler.request, _id=0)
            except Exception as e:
                url, status = seed.url, e.__class__.__name__
                msg = ''.join(traceback.format_exception(type(e), e, e.__traceback__))
                if getattr(e, "response", None) and isinstance(e.response, Res):
                    url = e.response.request.url
                    status = e.response.status_code
                self.spider_logging(seed, None, error=True, url=url, status=status, msg=msg)
                self.distribute(DealModel.fail, seed, _id=-1)

    def spider_logging(
            self, seed,
            item: Response = None,
            error: bool = False,
            dot: bool = True,
            **kwargs
    ):
        detail_log_info = LogTemplate.log_info(seed.to_dict)
        if error:
            url = kwargs.get("url")
            msg = kwargs.get("msg")
            status = kwargs.get("status")
            if dot:
                self.loghub_dot.build(
                    topic=urlparse(url).netloc,
                    data_size=-1, cost_time=-1,
                    status=status, url=url,
                    seed=seed.to_string,
                    proxy_type=seed.params.proxy_type,
                    proxy=seed.params.proxy,
                    project=self.project,
                    task=self.task, msg=msg,
                )
            logger.info(LogTemplate.download_exception.format(
                detail=detail_log_info,
                retry=seed.params.retry,
                priority=seed.params.priority,
                seed_version=seed.params.seed_version,
                identifier=seed.identifier or "",
                exception=msg
            ))
        else:
            logger.info(LogTemplate.download_info.format(
                detail=detail_log_info,
                retry=seed.params.retry,
                priority=seed.params.priority,
                seed_version=seed.params.seed_version,
                identifier=seed.identifier or "",
                status=item.response,
                response=LogTemplate.log_info(item.to_dict)
            ))
            if dot:
                end_time = time.time()
                stime = seed.params.start_time
                cost_time = end_time - stime if stime else -1
                topic = urlparse(item.response.request.url).netloc
                data_size = int(item.response.headers.get("content-length", 0))
                self.loghub_dot.build(
                    topic=topic, data_size=data_size, cost_time=cost_time,
                    status=200, seed=seed.to_string, url=item.response.url,
                    proxy=seed.params.proxy, proxy_type=seed.params.proxy_type,
                    project=self.project, task=self.task,
                )

    def run(self):
        self.register(self.loghub_dot.build_run, tag="LoghubDot")
        for _ in range(self.thread_num):
            self.register(self.spider, tag="Distributor")



import threading

from cobweb import setting
from typing import Callable
from cobweb.base import Queue
from abc import ABC, abstractmethod


class Scheduler(ABC, threading.Thread):

    __LAUNCHER_FUNC__ = ["_reset", "_scheduler", "_insert", "_refresh", "_delete"]

    def __init__(
            self,
            task,
            project,
            stop: threading.Event,
            pause: threading.Event,
            new: Queue,
            todo: Queue,
            done: Queue,
            upload: Queue,
            register: Callable
    ):
        super().__init__()
        self.task = task
        self.project = project

        self.task_model = setting.TASK_MODEL
        self.seed_reset_seconds = setting.SEED_RESET_SECONDS
        self.scheduler_wait_seconds = setting.SCHEDULER_WAIT_SECONDS
        self.new_queue_wait_seconds = setting.NEW_QUEUE_WAIT_SECONDS
        self.done_queue_wait_seconds = setting.DONE_QUEUE_WAIT_SECONDS
        self.todo_queue_full_wait_seconds = setting.TODO_QUEUE_FULL_WAIT_SECONDS
        self.before_scheduler_wait_seconds = setting.BEFORE_SCHEDULER_WAIT_SECONDS

        self.todo_queue_size = setting.TODO_QUEUE_SIZE
        self.new_queue_max_size = setting.NEW_QUEUE_MAX_SIZE
        self.done_queue_max_size = setting.DONE_QUEUE_MAX_SIZE
        self.upload_queue_max_size = setting.UPLOAD_QUEUE_MAX_SIZE

        self.stop = stop
        self.pause = pause

        self.new = new
        self.todo = todo
        self.done = done
        self.upload = upload

        self.register = register

        self.working_seeds = dict()

    def is_empty(self):
        if self.new.length == 0 and self.todo.length == 0 and self.done.length == 0 and self.upload.length == 0:
            return True
        else:
            return False

    def remove_working_seeds(self, seeds: list = None):
        for seed in seeds:
            if seed in self.working_seeds:
                self.working_seeds.pop(seed)

    @abstractmethod
    def reset(self):
        ...

    @abstractmethod
    def schedule(self):
        ...

    @abstractmethod
    def insert(self):
        ...

    @abstractmethod
    def refresh(self):
        ...

    @abstractmethod
    def delete(self):
        ...





# from .launcher_air import LauncherAir
# from .launcher_pro import LauncherPro
# from .launcher_api import LauncherApi


from .launcher import Launcher
from .uploader import Uploader
from .distributor import Distributor




import json
from typing import Union
from cobweb.base import (
    Seed,
    BaseItem,
    Request,
    Response,
    ConsoleItem,
)


class Crawler:

    @staticmethod
    def request(seed: Seed) -> Union[Request, BaseItem]:
        yield Request(seed.url, seed, timeout=5)

    @staticmethod
    def download(item: Request) -> Union[Seed, BaseItem, Response, str]:
        response = item.download()
        yield Response(item.seed, response, **item.to_dict)

    @staticmethod
    def parse(item: Response) -> BaseItem:
        upload_item = item.to_dict
        upload_item["text"] = item.response.text
        yield ConsoleItem(item.seed, data=json.dumps(upload_item, ensure_ascii=False))

